// BloodOmenDecompress - Decompresses .CTM and .CMP files from Blood Omen: Legacy of Kain
// Copyright 2010 Ben Lincoln
// http://www.thelostworlds.net/
//
// Decompression code based in part on portions of DeGob - GobEngine Script disassembler
// Copyright (C) 2008 The ScummVM project
// http://sourceforge.net/apps/trac/scummvm/browser/tools/trunk/degob_script.cpp?rev=40868

// This file is part of BloodOmenDecompress.

// BloodOmenDecompress is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BloodOmenDecompress is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BloodOmenDecompress (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;

namespace BloodOmenDecompress
{
    [StructLayout(LayoutKind.Explicit)]
    struct SixteenBitOverlay
    {
        [FieldOffset(0)]
        public short asShort;
        [FieldOffset(0)]
        public ushort asUShort;
    }

    class Program
    {
        protected const int _MaxIterations = 4096000;
        protected const long _MaxOutputSize = 67108864;

        static void Main(string[] args)
        {
            int resultCode = 0;

            if ((args.Length == 0) || (args.Length > 2))
            {
                Assembly currentAssembly = Assembly.GetExecutingAssembly();
                AssemblyName assemblyName = currentAssembly.GetName();
                int verMaj = assemblyName.Version.Major;
                int verMin = assemblyName.Version.Minor;
                int verBuild = assemblyName.Version.Build;

                Console.WriteLine("Blood Omen Decompressor");
                Console.WriteLine("Version " + verMaj.ToString() + "." + verMin.ToString() + "." + verBuild.ToString());
                Console.WriteLine("Copyright 2010 Ben Lincoln");
                Console.WriteLine("Legacy of Kain: The Lost Worlds");
                Console.WriteLine("http://www.thelostworlds.net/");
                Console.Write(Environment.NewLine);
                Console.WriteLine("Based in part on decompression code from DeGob - GobEngine Script disassembler (Copyright 2008 The ScummVM project).");
                Console.WriteLine("BloodOmenDecompress is distributed under version 3 of the GNU General Public License");
                Console.WriteLine("Please see the accompanying file LICENSE.txt for details");
                Console.Write(Environment.NewLine);
                Console.WriteLine("Usage: BloodOmenDecompress.exe INFILE [OUTFILE]");
                Console.Write(Environment.NewLine);
                Console.WriteLine("Where INFILE is a .CTM or .CMP file, and OUTFILE is optionally the output filename.");
                Console.WriteLine("If no output filename is specified, the program will default to the same name as the input, substituting the extension .TIM for .CTM input files, .MAP for .CMP input files, and .UNC for all other extensions.");
                resultCode = 1;
            }
            else
            {
                string inFilePath = args[0];
                string outFilePath = "";
                if (args.Length == 2)
                {
                    outFilePath = args[1];
                }
                else
                {
                    string currentExt = Path.GetExtension(inFilePath).ToUpper().Trim().Replace(".", "");
                    string newExt = "UNC";
                    switch (currentExt)
                    {
                        case "CTM":
                            newExt = "TIM";
                            break;
                        case "CMP":
                            newExt = "MAP";
                            break;
                        default:
                            newExt = "UNC";
                            break;
                    }
                    outFilePath = Path.ChangeExtension(inFilePath, newExt);
                }

                if (!File.Exists(inFilePath))
                {
                    Console.WriteLine("Unable to locate the input file '" + inFilePath + "'.");
                    resultCode = 1;
                }
                else
                {
                    try
                    {
                        resultCode = DecompressFile(inFilePath, outFilePath);
                    }
                    catch (Exception ex)
                    {
                        resultCode = 0x1911;
                        Console.WriteLine("An unexpected error has occurred. The exception description was '" + ex.Message + "'. If this occurred while decompressing a legitimate Blood Omen CTM or CMP file, please notify the developer.");
                    }
                }
            }

            System.Environment.ExitCode = resultCode;
        }

        protected static int DecompressFile(string inPath, string outPath)
        {
            FileStream inStream = new FileStream(inPath, FileMode.Open, FileAccess.Read);
            FileStream outStream = new FileStream(outPath, FileMode.Create, FileAccess.ReadWrite);
            BinaryReader inReader = new BinaryReader(inStream);
            BinaryWriter outWriter = new BinaryWriter(outStream);
            BinaryReader outReader = new BinaryReader(outStream);

            inStream.Seek(0, SeekOrigin.Begin);

            FileInfo inInfo = new FileInfo(inPath);
            long compressedSize = inReader.ReadUInt32() + 4;
            if (compressedSize > inInfo.Length)
            {
                Console.WriteLine("Error: the file '" + inPath + "' has a header indicating it should be " +
                    compressedSize.ToString() + " bytes in size, but the actual file size is " + inInfo.Length.ToString() +
                    " bytes. The file is either incomplete or is not a compressed Blood Omen data file.");
                return 10;
            }
            if (compressedSize < inInfo.Length)
            {
                Console.WriteLine("Warning: the file '" + inPath + "' has a header indicating it should be " +
                    compressedSize.ToString() + " bytes in size, but the actual file size is " + inInfo.Length.ToString() +
                    " bytes. The file either contains extra data or is not a compressed Blood Omen data file.");
            }

            bool doneDecompressing = false;

            int numIterations = 0;
            int bytesWritten = 0;

            ushort tempIndex = 4078;
            ushort command = 0;
            short offset;
            byte length;

            byte[] tempBuffer = new byte[4114];
            for (int i = 0; i < 4078; i++)
            {
                tempBuffer[i] = 0x20;
            }

            while (!doneDecompressing)
            {
                if (numIterations > _MaxIterations)
                {
                    Console.WriteLine("Error: the program has iterated " + numIterations.ToString() + " times, " + 
                        "which is greater than the maximum number permitted. This indicates that the input is not " + 
                        "a compressed Blood Omen data file.");
                    return 21;
                }
                if (outStream.Position > _MaxOutputSize)
                {
                    Console.WriteLine("Error: the output file size has reached " + outStream.Position.ToString() + " bytes, " + 
                        "which is greater than the maximum size permitted. This indicates that the input is not " + 
                        "a compressed Blood Omen data file.");
                    return 22;
                }

                command = (ushort)(command >> 1);

                if ((command & 0x0100) == 0)
                {
                    command = (ushort)((ushort)(inReader.ReadByte()) | 0xFF00);
                }
                    if ((command & 0x01) == 0x01)
                    {
                        byte currentByte = inReader.ReadByte();
                        outWriter.Write(currentByte);
                        bytesWritten++;
                        tempBuffer[tempIndex] = currentByte;
                        tempIndex++;
                        //tempIndex = (ushort)(tempIndex % 4096);
                        tempIndex = (ushort)(tempIndex & 0x0FFF);   // may be slightly faster?
                        if (inStream.Position >= inStream.Length)
                        {
                            break;
                        }
                    }
                    else
                    {
                        byte currentByte1 = inReader.ReadByte();
                        byte currentByte2 = inReader.ReadByte();
                        // Some sleight of hand is required here because of how C# handles casting between 
                        // signed and unsigned data types of the same length
                        // The data should never result in the sign bit being negative, but just in case it 
                        // does...
                        int offsetTemp = currentByte1;
                        offsetTemp |= (currentByte2 & 0xF0) << 4;
                        SixteenBitOverlay offsetOL = new SixteenBitOverlay();
                        offsetOL.asUShort = (ushort)offsetTemp;
                        offset = offsetOL.asShort;

                        length = (byte)((currentByte2 & 0x0F) + 3);

                        for (int i = 0; i < length; i++)
                        {
                            //outWriter.Write(tempBuffer[(offset + i) % 4096]);
                            outWriter.Write(tempBuffer[(offset + i) & 0x0FFF]);   // may be slightly faster?
                            bytesWritten++;
                            if (inStream.Position >= inStream.Length)
                            {
                                doneDecompressing = true;
                                break;
                            }
                            //tempBuffer[tempIndex] = tempBuffer[(offset + i) % 4096];
                            tempBuffer[tempIndex] = tempBuffer[(offset + i) & 0x0FFF];   // may be slightly faster?
                            tempIndex++;
                            //tempIndex = (ushort)(tempIndex % 4096);
                            tempIndex = (ushort)(tempIndex & 0x0FFF);   // may be slightly faster?
                        }
                    }
                numIterations++;
            }

            // Playstation files apparently need to be multiples of 1024 bytes in size
            // (or maybe it's just PSXMC that complains if they aren't).
            if ((bytesWritten % 1024) > 0)
            {
                int numPadBytes = 1024 - (bytesWritten % 1024);
                for (int i = 0; i < numPadBytes; i++)
                {
                    outWriter.Write((byte)0x00);
                }
            }

            outStream.Close();
            inStream.Close();

            return 0;
        }
    }
}
